//
//  CGOLModel.swift
//  ConwaysGameOfLife
//
//  Created by narita on 2018/08/22.
//  Copyright © 2018年 narita. All rights reserved.
//

import Foundation

// LifeGameの碁盤を定義するための構造体の定義

// 場所
struct CGOLPoint
{
    var x: NSInteger = 0
    var y: NSInteger = 0
    
    init(x: NSInteger, y: NSInteger)
    {
        self.x = x
        self.y = y
    }
}

// 広さ
struct CGOLSize
{
    var width : NSInteger = 0
    var height: NSInteger = 0
    
    init(width: NSInteger, height: NSInteger)
    {
        self.width = width
        self.height = height
    }
}

// 範囲
struct CGOLRect
{
    var origine: CGOLPoint = CGOLPoint(x:0, y:0)
    var size   : CGOLSize  = CGOLSize(width:0, height:0)
    
    init(x: NSInteger, y: NSInteger,
         width: NSInteger, height: NSInteger)
    {
        self.origine.x = x
        self.origine.y = y

        self.size.width = width
        self.size.height = height
    }

    init(origine: CGOLPoint, size: CGOLSize)
    {
        self.origine = origine
        self.size    = size
    }
}

// 生死
enum CGOLBool: NSInteger
{
    case died  = 0
    case alive = 1
}

// 状態
enum CGOLState: NSInteger
{
    case stop    = 0
    case start   = 1
    case running = 2
}

// 通知用定数の定義
extension Notification.Name
{
    // モデルが次の状態になった場合の通知
    static let cellsUpdateNotification = Notification.Name("cellsUpdate")

    // サイズが変更された場合の通知
    static let sizeChangeNotification = Notification.Name("sizeChange")
}

class CGOLModel: NSObject
{
    // MARK: - class methods
    
    // MARK: - init methods
    
    override convenience init()
    {
        // デフォルトのサイズ
        self.init(width: 64, height: 64)
    }
    
    convenience init(size: CGOLSize)
    {
        self.init(width: size.width, height: size.height)
    }

    init(width: NSInteger, height: NSInteger)
    {
        self.cellsSize = CGOLSize(width: width, height: height)

        super.init()

        self.cells = NSMutableArray(capacity:
            (self.cellsSize.width + 2) * (self.cellsSize.height + 2) )
        
        self.randomizeAll()
    }
    
    // MARK: - dealloc
    
    // MARK: - NSCopying, hash, isEqual:
    
    // MARK: - NSCoding methods
    
    func writeToData() throws -> Data
    {
        // プロパティリスト形式のファイルを吐き出す
        let theDict: NSMutableDictionary = NSMutableDictionary()
        
        // 大きさの保存
        let w: NSInteger = self.cellsSize.width
        let h: NSInteger = self.cellsSize.height
        
        theDict.setObject(NSNumber(value:w), forKey: "width" as NSCopying)
        theDict.setObject(NSNumber(value:h), forKey: "height" as NSCopying)
        
        // 配列の作成と保存
        // わざわざ配列を作り直しているのは、Modelの内部構造への依存を排除するため
        let theArray: NSMutableArray = NSMutableArray.init(capacity: w * h)
        
        for y in 0 ..< h
        {
            for x in 0 ..< w
            {
                let thePoint = CGOLPoint(x: x, y: y)
                let theValue = self.alive(at: thePoint)
                let theObject = NSNumber(integerLiteral: theValue.rawValue)
                theArray.add(theObject)
            }
        }
        
        theDict.setObject(theArray, forKey: "matrix" as NSCopying)
        
        let theData = try PropertyListSerialization.data(fromPropertyList: theDict,
                                                         format: .xml,
                                                         options: 0)
        
        return theData as Data
    }
    
    func read(from data: Data) throws
    {
        let theDict = try PropertyListSerialization.propertyList(from: data,
                                                                 options:[],
                                                                 format:nil) as! NSDictionary

        let w: NSInteger = ( theDict.object(forKey: "width") as! NSNumber ).intValue
        let h: NSInteger = ( theDict.object(forKey: "height") as! NSNumber ).intValue
        
        let theArray: NSArray = theDict.object(forKey: "matrix") as! NSArray
        
        self.cellsSize = CGOLSize(width: w, height: h)

        for y in 0 ..< h
        {
            for x in 0 ..< w
            {
                let thePoint = CGOLPoint(x: x, y: y)
                let theObject = theArray[y * w + x] as! NSNumber
                
                self.setAlive(CGOLBool(rawValue: theObject.intValue)!, at: thePoint)
            }
        }

        NotificationCenter.default.post(name: .cellsUpdateNotification,
                                        object: self)
    }
    
    // MARK: - restorableState methods
    
    // MARK: - life cycle methods
    
    // MARK: - action methods
    
    func clearAll()
    {
        self.cells.removeAllObjects()
        
        let w = self.cellsSize.width
        let h = self.cellsSize.height

        let theLength = (w + 2) * (h * 2)
        
        for _: NSInteger in 0 ..< theLength
        {
            let theValue: NSNumber! = NSNumber(value: CGOLBool.died.rawValue)
            
            self.cells.add(theValue)
        }
        
        NotificationCenter.default.post(name: .cellsUpdateNotification,
                                        object: self)
    }
    
    func randomizeAll()
    {
        self.clearAll()
        
        let w = self.cellsSize.width
        let h = self.cellsSize.height
        
        for y: NSInteger in 0 ..< h
        {
            for x: NSInteger in 0 ..< w
            {
                let theValue: CGOLBool
                theValue = arc4random() % 2 == 0 ? CGOLBool.died : CGOLBool.alive

                let thePoint = CGOLPoint(x: x, y: y)

                self.setAlive(theValue, at: thePoint)
            }
        }
    }
    
    func alive(at point: CGOLPoint) -> CGOLBool
    {
        // 境界分だけズラす
        let x = point.x + 1
        let y = point.y + 1
        let w = self.cellsSize.width
        
        let theValue: NSNumber! = self.cells.object(at: y * w + x) as! NSNumber
        
        return CGOLBool(rawValue: theValue.intValue)!
    }
    
    func setAlive(_ alive :CGOLBool, at point: CGOLPoint)
    {
        // 境界分だけズラす
        let x = point.x + 1
        let y = point.y + 1
        let w = self.cellsSize.width
        
        let theValue: NSNumber! = NSNumber(value: alive.rawValue)
        
        self.cells.replaceObject(at: y * w + x, with: theValue)
        
        // viewへ updateされた事を通知
        NotificationCenter.default.post(name: .cellsUpdateNotification,
                                        object: self)

    }
    
    func start()
    {
        // 停止中ならば
        if self.state == .stop
        {
            // 状態を実行中にして
            self.state = .start
            
            // timerをセット
            self.timer = Timer.scheduledTimer(timeInterval: 0.3,
                                              target: self,
                                              selector: #selector(next),
                                              userInfo: nil,
                                              repeats: false)
        }
    }

    func stop()
    {
        // 実行中ならば
        // 状態を停止要求に変更
        if self.state == .running
        {
            self.state = .stop
        }
    }

        
    @objc func next()
    {
        if self.state == .start || self.state == .running
        {
            
            
            let theNextCells: NSMutableArray = self.cells.mutableCopy() as! NSMutableArray
            
            let w = self.cellsSize.width
            let h = self.cellsSize.height
            
            for y: NSInteger in 0 ..< h
            {
                for x: NSInteger in 0 ..< w
                {
                    let thePoint = CGOLPoint(x: x, y: y)
                    
                    let theCurrentValue: CGOLBool = self.alive(at: thePoint)
                    let theNeighValue: NSInteger = self.sumAliveNeighborCells(at: thePoint)
                    var theNextValue: CGOLBool    = theCurrentValue
                    
                    if theCurrentValue == .alive
                    {
                        if( !((theNeighValue == 2) || (theNeighValue == 3)) )
                        {
                            theNextValue = .died;
                        }
                    }
                    else
                    {
                        if( theNeighValue == 3 )
                        {
                            theNextValue = .alive;
                        }
                    }
                    
                    // 結果の書き込み
                    do
                    {
                        // 境界分だけズラす
                        let x = thePoint.x + 1
                        let y = thePoint.y + 1
                        let w = self.cellsSize.width
                        
                        let theValue: NSNumber! = NSNumber(value: theNextValue.rawValue)
                        
                        theNextCells.replaceObject(at: y * w + x, with: theValue)
                    }
                }
            }
            
            self.cells = theNextCells
            
            // viewへ updateされた事を通知
            NotificationCenter.default.post(name: .cellsUpdateNotification,
                                            object: self)
            
            self.state = .running

            // timerを再びセット
            self.timer = Timer.scheduledTimer(timeInterval: 0.3,
                                              target: self,
                                              selector: #selector(next),
                                              userInfo: nil,
                                              repeats: false)

        }
    }
    
    private func sumAliveNeighborCells(at point: CGOLPoint) -> NSInteger
    {
        var theResult: NSInteger = 0
        
        // 上段
        if self.alive(at: CGOLPoint(x: point.x - 1, y: point.y - 1)) == .alive
        { theResult += 1 }
        
        if self.alive(at: CGOLPoint(x: point.x ,    y: point.y - 1)) == .alive
        { theResult += 1 }
        
        if self.alive(at: CGOLPoint(x: point.x + 1, y: point.y - 1)) == .alive
        { theResult += 1 }

        // 中段
        if self.alive(at: CGOLPoint(x: point.x - 1, y: point.y    )) == .alive
        { theResult += 1 }
        
        if self.alive(at: CGOLPoint(x: point.x + 1, y: point.y    )) == .alive
        { theResult += 1 }

        // 下段
        if self.alive(at: CGOLPoint(x: point.x - 1, y: point.y + 1)) == .alive
        { theResult += 1 }
        
        if self.alive(at: CGOLPoint(x: point.x ,    y: point.y + 1)) == .alive
        { theResult += 1 }
        
        if self.alive(at: CGOLPoint(x: point.x + 1, y: point.y + 1)) == .alive
        { theResult += 1 }

        return theResult
    }
    
    // MARK: - event handling methods
    
    // MARK: - drawing methods
    
    // MARK: - delegate/datasource methods
    
    // MARK: - accessor methods (in pairs)
    
    // 実行させるためのタイマー
    var timer: Timer? = nil
    
    // 実行の状態
    var state: CGOLState = .stop

    // 大きさ
    var cellsSize: CGOLSize = CGOLSize(width: 0, height: 0)
    {
        willSet(newSize)
        {
            
        }
        
        didSet
        {
            self.cells = NSMutableArray(capacity:
                (self.cellsSize.width + 2) * (self.cellsSize.height + 2) )
            
            self.clearAll()
            
            NotificationCenter.default.post(name: .sizeChangeNotification,
                                            object: self)
        }
    }

    // 上下左右に1列を追加することで、次の状態を計算する時の境界条件分岐を省略する。
    private var cells: NSMutableArray = NSMutableArray()
    
    // MARK: - Utility methods
    
    override var description: String
    {
        get
        {
            let theResult = NSMutableString()
            
            let w = self.cellsSize.width
            let h = self.cellsSize.height
            
            for y: NSInteger in 0 ..< h
            {
                for x: NSInteger in 0 ..< w
                {
                    let thePoint = CGOLPoint(x: x, y: y)

                    let theValue: CGOLBool = self.alive(at: thePoint)
                    
                    switch theValue
                    {
                        case .alive:
                            theResult.append("O")
                        case .died:
                            theResult.append(".")
                    }
                }
                
                theResult.append("\r")
            }
            
            return theResult as String
        }
    }
    
    // descriptionで出力した形式を読み込む。エラー処理などは加えていないテスト用です。
    /* private */func gameBoardSetup(newLifeGameBoard: NSString!)
    {
        let w = self.cellsSize.width
        let h = self.cellsSize.height

        var theRange: NSRange = NSRange(location: 0, length: 1)
        
        for y: NSInteger in 0 ..< h
        {
            for x: NSInteger in 0 ..< w
            {
                let thePoint = CGOLPoint(x: x, y: y)
                
                let theSubString: NSString = newLifeGameBoard.substring(with: theRange) as NSString

                switch theSubString
                {
                case "O":
                    self.setAlive(.alive, at: thePoint)
                case ".":
                    self.setAlive(.died, at: thePoint)
                default:
                    assert(false)
                }

                theRange.location += 1
            }
            
            let theSubString: NSString = newLifeGameBoard.substring(with: theRange) as NSString
            
            if theSubString == "\r"
            {
                // 改行
            }
            else
            {
                assert(false)
            }
            theRange.location += 1
        }
    }
}
